unit usmoke;

// __PTC_RESIZE_WINDOW__        Resizable window
// __PTC_ALLOW_CLOSE__          window can be closed
// __PTC_CLOSE_ON_ESCAPE__      Close window when escape is pressed
// __PTC_WINDOWED__
// __PTC_SYSTEM_MENU__
// __PTC_CENTER_WINDOW__
// __PTC_DISABLE_SCREENSAVER__
// __PTC_ICON__                 Show Icon
// __PTC_DDRAW__
// __PTC_GDI__
// __PTC_VFW__

interface

procedure MainFunc;

implementation

uses
  mmsystem,
  windows,
  messages,
{$ifdef __PTC_VFW__}
  vfw8;
{$endif}
{$ifdef __PTC_GDI__}
  gdi8;
{$endif}

type
  PByteA = ^TByteA;
  TByteA = array [Word] of Byte;
  PLongA = ^TLongA;
  TLongA = array [0..0] of LongWord;
  TQuad = packed record
  case integer of
    0: (Color: Longint);
    1: (b1,b2,b3,b4: byte);
  end;

var
  Width:  Integer;
  Height: Integer;
  Size:   Integer;
  pixel: array of Byte;
  map: array of array of Byte;
  Pal: array[0..255] of TRGBQUAD;
  st2: longword;

const
  B:  Integer = $100;
  BM: Integer = $FF;
  N:  Integer = $1000;
var
  P:  array[0..$202] of Integer;
  G1: array[0..$202] of Double;
  G2i: array[0..$202,0..1] of Integer;

  Pointbmp: array[0..256] of Byte;

const
  nrpoints = 256;
  nrlines = 32;

procedure BuildPointsbmp;
var
  x,y,t: Integer;
begin
  for x := 0 to 16 do
    for y := 0 to 16 do begin
       t := sqr(x - 8) + sqr(y - 8);
       if t > 64 then
         Pointbmp[x + 16 * y] := 0
       else
         Pointbmp[x + 16 * y] := round(3 * (8 - sqrt(t)));
    end;
end;

function FloatToStr(Value: Extended): string;
begin
  Result := '';
  if Value > 1000 then
    Result := Result + Char(Ord('0') + Trunc(Value/1000));
  if Value > 100 then begin
    Value := value - 1000 * Trunc(Value/1000);
    Result := Result + Char(Ord('0') + Trunc(Value/100));
  end;
  if Value > 10 then begin
    Value := value - 100 * Trunc(Value/100);
    Result := Result + Char(Ord('0') + Trunc(Value/10));
  end;
  Value := value - 10 * Trunc(Value/10);
  Result := Result + Char(Ord('0') + Trunc(Value)) + '.';
  Value := value - Trunc(Value);
  Value := Value * 10;
  Result := Result + Char(Ord('0') + Trunc(Value));
  Value := value - Trunc(Value);
  Value := Value * 10;
  Result := Result + Char(Ord('0') + Trunc(Value));
end;

procedure CreatePalette;
var
  i: integer;
begin
  for i := 0 to 255 do begin
    Pal[i].rgbred := i;
    Pal[i].rgbgreen := i;
    Pal[i].rgbBlue := 64 + round(191 * sqrt(i/255));
    Pal[i].rgbreserved := i;
  end;
end;

function noise1D(const x: double): double;
var
  bx0,bx1: integer;
  rx0,rx1,sx,t,u,v: double;
begin
  t := x + N;
  bx0 := Trunc(t) and BM;
  bx1 := (bx0 + 1) and BM;
  rx0 := t - trunc(t);
  rx1 := rx0 - 1;

  sx := rx0 * rx0 * (3 - 2 * rx0);
  u  := rx0 * G1[P[bx0]];
  v  := rx1 * G1[P[bx1]];
  result := u + sx * (v - u);
end;

function turbulence1D(x: double; const n: integer): double;
var
  freq: double;
  i: integer;
begin
  result := 0;
  freq := 1.0;

  for i := n - 1 downto 0 do begin
    result := result + Noise1D(x) * freq;
    x := x * 2;
    freq := freq * 0.5;
  end;
end;

procedure initnoise(seed: Integer);
var
  I,J,T: integer;
  len: double;
  G2: array[0..1] of double;
begin
  randseed := seed;
  if seed = 0 then
    randomize;

  for i := 0 to B - 1 do begin
    P[i] := i;

		G1[i] := (Trunc(Random * 2 * B) - B)/B;

		G2[0] := (Trunc(Random * 2 * B) - B)/B;
		G2[1] := (Trunc(Random * 2 * B) - B)/B;
    len := sqrt(G2[0] * G2[0] + G2[1] * G2[1]);
    if len > 1E-5 then begin
      G2[0] := G2[0] / len;
      G2[1] := G2[1] / len;
    end;

		G2i[i,0] := trunc(G2[0] * 1024);
		G2i[i,1] := trunc(G2[1] * 1024);
  end;

  for i := 0 to B - 1 do begin
    j := Trunc(Random * B);
    T := P[i];
    P[i] := P[j];
    P[j] := T;
  end;

  for i := 0 to B + 1 do begin
    P[B + i] := P[i];

    G1[B + i] := G1[i];

    G2i[B + i][0] := G2i[i][0];
    G2i[B + i][1] := G2i[i][1];
  end;
end;

function noise2Di(const x,y: Integer): Integer;
var
  bx0,bx1,by0,by1: integer;
  b00,b10,b01,b11: integer;
  rx0, rx1, ry0, ry1: integer;
  sx, sy,t,a,b,u,v: integer;
  i,j: integer;
  r: integer;
begin
  t := x + $40000000;
  bx0 := (t shr 10) and $FF;
  bx1 := (bx0 + 1) and $FF;
  i := P[bx0];
  rx0 := t and 1023;
  rx1 := rx0 - 1024;

  t := y + $40000000;
  by0 := (t shr 10) and $FF;
  by1 := (by0 + 1) and $FF;
  j := P[bx1];
  ry0 := t and 1023;
  ry1 := ry0 - 1024;

  b00 := P[i + by0];
  b10 := P[j + by0];
  b01 := P[i + by1];
  b11 := P[j + by1];

  sx := (rx0 * rx0 * (3072 - 2 * rx0)) shr 20;
  sy := (ry0 * ry0 * (3072 - 2 * ry0)) shr 20;

  u := (rx0 * G2i[b00][0] + ry0 * G2i[b00][1]);
  v := (rx1 * G2i[b10][0] + ry0 * G2i[b10][1]);
  a := (u shl 10) + sx * (v - u);
asm
  SAR a,20
end;

  u := (rx0 * G2i[b01][0] + ry1 * G2i[b01][1]);
  v := (rx1 * G2i[b11][0] + ry1 * G2i[b11][1]);
  b := (u shl 10)  + sx * (v - u);
asm
  SAR b,20
end;

  r := a shl 10  + sy * (b - a);
asm
  SAR r,10
end;

  result := r;
end;

function turbulence2Di(x,y: Integer; const n: integer): Integer;
var
  r, i: integer;
  a: integer;
begin
  r := 0;
  a := 1;

  for i := n - 1 downto 0 do begin
    Inc(r, Noise2Di(x,y) div a);
    x := x shl 1;
    y := y shl 1;
    a := a shl 1;
  end;
  Result := r;
end;

procedure BuildPicture;
var
  i,j: longint;
  t: longint;
  ix,iy: integer;
  wind : Integer;
  p: longint;
  col: byte;
  ifacx, ifacy: array[0..nrpoints] of integer;
  jfacx, jfacy: array[0..nrlines] of integer;
begin
  fillchar(Pixel[0], size, 0);

  t := (timeGetTime - St2) div 4;

  wind := Round(500 * turbulence1D(t/8000, 2));

  for i := 0 to nrpoints do begin
    ifacx[i] := t + round(2500 * (nrpoints - i)/nrpoints);
    ifacy[i] := t + round(3000 * (nrpoints - i)/nrpoints) + 3000;
  end;

  for j := 0 to nrlines do begin
    jfacx[j] := round(600 * j/nrlines);
    jfacy[j] := round(700 * j/nrlines);
  end;

  for i := nrpoints downto 0 do begin
    col := round(255 * (nrpoints - i)/nrpoints);
    for j := nrlines downto 0  do begin
//      x := (turbulence2Di(ifacx[i], jfacx[j], 3) + wind) * i/nrpoints/ 1024;
//      y := (turbulence2Di(ifacy[i], jfacy[j], 2) + 1024) * i/nrpoints/ 1024;
      ix := (turbulence2Di(ifacx[i], jfacx[j], 3) + wind) * i * Height;
      iy := (turbulence2Di(ifacy[i], jfacy[j], 2) + 1000) * 3 * i * Height;
asm
  SAR ix, 18
  SAR iy, 20
end;

      ix := Width div 2 + ix;
      iy := Height - iy;

      p := ix + Width * iy;
      if (p >= 0) and (p < size) then
        Pixel[p] := col;
    end;
  end;
end;

procedure BuildPictureWU;
var
  i,j: longint;
  t: longint;
  ix,iy: integer;
  wind : Integer;
  p: longint;
  ifacx, ifacy: array[0..nrpoints] of integer;
  jfacx, jfacy: array[0..nrlines] of integer;
  pp : PByteA;
  u,v: integer;
  c: integer;
  d: integer;
begin
  fillchar(Pixel[0], size, 0);

  t := (timeGetTime - St2) div 4;

  wind := Round(500 * turbulence1D(t/8000, 2));

  for i := 0 to nrpoints do begin
    ifacx[i] := t + round(2500 * (nrpoints - i)/nrpoints);
    ifacy[i] := t + round(3000 * (nrpoints - i)/nrpoints) + 3000;
  end;

  for j := 0 to nrlines do begin
    jfacx[j] := round(600 * j/nrlines);
    jfacy[j] := round(700 * j/nrlines);
  end;

  for i := nrpoints downto 0 do begin
    c := 255 - round(255 * i/nrpoints);
    for j := nrlines downto 0 do begin
      ix := (turbulence2Di(ifacx[i], jfacx[j], 3) + wind) * i * Height;
      iy := (turbulence2Di(ifacy[i], jfacy[j], 2) + 1000) * 3 * i * Height;
asm
  SAR ix, 10
  SAR iy, 12
end;
      u := (ix + 1000000) and 255;
      v := (iy + 1000000) and 255;

      ix := (Width * 128 + ix) div 256;
      iy := Height - iy div 256;

      p := ix + Width * iy;
      if (p >= 0) and (p < (size - width - 1)) then begin
        pp := @Pixel[p];
        d := (c * (255 - u) * (255 - v)) shr 16;
        d := d + pp[0];
        if d > 255 then d := 255;
        pp[0] := d;

        inc(Integer(pp));
        d := (c *        u  * (255 - v)) shr 16;
        d := d + pp[0];
        if d > 255 then d := 255;
        pp[0] := d;

        inc(Integer(pp), Width - 1);
        d := (c * (255 - u) *        v ) shr 16;
        d := d + pp[0];
        if d > 255 then d := 255;
        pp[0] := d;

        inc(Integer(pp));
        d := (c *        u  *        v) shr 16;
        d := d + pp[0];
        if d > 255 then d := 255;
        pp[0] := d;
      end;
    end;
  end;
end;

procedure BuildPictureBlob;
var
  i,j: longint;
  t: longint;
  ix,iy: integer;
  x,y: integer;
  wind : Integer;
  p: longint;
  c, d: integer;
  ifacx, ifacy: array[0..nrpoints] of integer;
  jfacx, jfacy: array[0..nrlines] of integer;
  pp: PByteA;
  pb: Integer;
begin
  fillchar(Pixel[0], size, 0);

  t := (timeGetTime - St2) div 3;

  wind := Round(500 * turbulence1D(t/8000, 2));

  for i := 0 to nrpoints do begin
    ifacx[i] := t + round(2500 * (nrpoints - i)/nrpoints);
    ifacy[i] := t + round(3000 * (nrpoints - i)/nrpoints) + 3000;
  end;

  for j := 0 to nrlines do begin
    jfacx[j] := round(600 * j/nrlines);
    jfacy[j] := round(700 * j/nrlines);
  end;

  for i := nrpoints downto 0 do begin
    c := 255 - round(255 * i/nrpoints);
    for j := nrlines downto 0  do begin
      ix := (turbulence2Di(ifacx[i], jfacx[j], 3) + wind) * i * Height;
      iy := (turbulence2Di(ifacy[i], jfacy[j], 2) + 1000) * 3 * i * Height;
asm
  SAR ix, 18
  SAR iy, 20
end;
      ix := Width div 2 + ix;
      iy := Height - iy - 16;

      p := ix + Width * iy;
      if (p >= 0) and (p < size - 16 * width) then begin
        pp := @Pixel[p];
        pb := 0;
        for y := 15 downto 0 do begin
          for x := 15 downto 0 do begin
            d := (c * Pointbmp[pb]) shr 8;
            d := d + pp[0];
            if d > 255 then d := 255;
            pp[0] := d;

            Inc(Integer(pp));
            Inc(pb);
          end;
          Inc(Integer(pp), width - 16);
        end;
      end;
    end;
  end;
end;

procedure MainFunc;
{$ifdef __PTC_WINDOWED__}
var
  NrFrame:   Integer;
  t, startTime: double;
  s: string;
{$endif}
begin
{$ifdef __PTC_WINDOWED__}
  startTime := timeGetTime * 0.001;
  NrFrame := 0;
{$endif}
  Width := 640;
  Height:= 480;
  Size  := Height * Width;
  SetLength(pixel, Size);
  SetLength(Map, Height + 1, Width + 1);

  st2 := timeGetTime;

  CreatePalette;
  BuildPointsbmp;
  InitNoise(0);

  if (ptc_open('Smoke', Width, Height) = 0) then
    Exit;

  ptc_setpalette(pal);
  while (True) do begin
    BuildPictureBlob;

    ptc_update(@pixel[0]);

{$ifdef __PTC_WINDOWED__}
    Inc(NrFrame);
    t := (timeGetTime * 0.001) - StartTime;
    if t > 2 then begin
      s := 'Smoke ' + FloatToStr(NrFrame/t) + ' fps';

      SendMessage(wnd, WM_SETTEXT, 0, Longint(Pchar(s)));

      startTime := timeGetTime * 0.001;
      NrFrame := 0;
    end;
{$endif}
  end;
  ptc_close;
end;

end.
